#include <sstream>
#include <iostream>
#include <string>
#include <deque>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/epoll.h>

#include <protocol.h>

#define max(a,b)    ((a) > (b) ? (a) : (b))

static int CLIENTS_NUMB=4;
static int packages_delivered=0;
static unsigned long long bytes_delivered=0;
int epfd=0;

/* {{{ all_read */
ssize_t all_read(int fd, void *buf, size_t count)
{
	size_t bytes_done=0, res, to_read;
	while(bytes_done < count) {
		if( (count-bytes_done) >16000) {
			to_read=16000;
		} else {
			to_read=count-bytes_done;
		}

		res = read(fd, (unsigned char *)buf+bytes_done, to_read);
		if(res<0) {
			printf("SWFCOM_ERROR: read error\n");
			exit(1);
		}
		bytes_done+=res;
	}
	return bytes_done;
}
/* }}} */

/* {{{ all_write */
ssize_t all_write(int fd, const void *buf, size_t count)
{
	size_t bytes_done=0, res, to_write;
	while(bytes_done < count) {
		if( (count-bytes_done) >16000) {
			to_write=16000;
		} else {
			to_write=count-bytes_done;
		}
		res = write(fd, (unsigned char *)buf+bytes_done, to_write);
		if(res<0) {
			printf("SWFCOM_ERROR: write did not succeed\n");
			exit(1);
		}
		bytes_done+=res;
	}
	return bytes_done;
}
/* }}} */

/* the normal Packet class it a bit to bulky and lacks a field for the data */
/* {{{ class mini_Packet */
class mini_Packet {
    public:
        unsigned to;
        unsigned from;
        unsigned size;
        unsigned tag;
        char *data;
        mini_Packet(unsigned to_, unsigned from_, unsigned size_, unsigned tag_);
        mini_Packet();
        ~mini_Packet();
};
mini_Packet::mini_Packet()
{
}
mini_Packet::mini_Packet(unsigned to_, unsigned from_, unsigned size_, unsigned tag_)
{
	to = to_;
	from = from_;
	size = size_;
	tag = tag_;
	data = NULL;
}
mini_Packet::~mini_Packet()
{
}
/* }}} class mini_Packet */

enum ClientState {
	NOTHING,   // as in nothing todo
	RECV_WAIT, // we are waiting for a package
	FINISHED   // connection closed, finished with this node
};

/* we need an "unique" tag so that we can find the corresponding data to a package */
unsigned unique_tag_state;

typedef struct ClientControlBlock {
	int fd;

	unsigned long long time;
	ClientState state;
	std::deque<mini_Packet> incomming_packets;
	} ClientControlBlock;
ClientControlBlock* Clients;

/* {{{ HandleClient */
void HandleClient(unsigned clientid)
{
	switch(Clients[clientid].state) {
		case RECV_WAIT: {
			DataMessage md;
			if(Clients[clientid].incomming_packets.size() > 0) {
				std::cout << "There is a packet for " << clientid << std::endl;
				md.request=SWORDFISH_RECV;
				md.param1=Clients[clientid].incomming_packets.front().from;
				md.param2=Clients[clientid].incomming_packets.front().tag;
				md.param3=Clients[clientid].incomming_packets.front().size;
				write(Clients[clientid].fd,&md,sizeof(unsigned int)*4);
				all_write(Clients[clientid].fd,Clients[clientid].incomming_packets.front().data,md.param3);
				free(Clients[clientid].incomming_packets.front().data);
				Clients[clientid].incomming_packets.pop_front();
				Clients[clientid].state = NOTHING;
				packages_delivered++;
				bytes_delivered+=md.param3;
			}
			break;
		}
		default: {
			break;
		}
	}
}
/* }}} */

/* {{{ Worker */
void Worker()
{
	for(int i=0; i < CLIENTS_NUMB; i++)
		if(Clients[i].state == RECV_WAIT)
			HandleClient(i);
}
/* }}} */

/* {{{ HandleMsg */
void HandleMsg(int clientid)
{
	DataMessage md;
	read(Clients[clientid].fd,&md,6*4);
//	std::cout << "HandleMsg: " << clientid << " request: " << md.request << std::endl;
	switch(md.request) {
		case SWORDFISH_FINISH: {
			Clients[clientid].state = FINISHED;
			epoll_ctl(epfd, EPOLL_CTL_DEL, Clients[clientid].fd, NULL);
			close(Clients[clientid].fd);
			int all_finished=1;
			for(int i=0; i < CLIENTS_NUMB; i++) {
				if(Clients[i].state!=FINISHED) all_finished=0;
			}
			if(all_finished==1) {
				std::cout << "finished" << std::endl;
				std::cout << "packages delivered overall: " << packages_delivered << std::endl;
				std::cout << "bytes delivered overall: " << bytes_delivered << std::endl;
				exit(0);
			}
			break;
		}
		case SWORDFISH_SIZE: {
			md.param1 = CLIENTS_NUMB;
			std::cout << "size " << md.param1 << std::endl;
			write(Clients[clientid].fd,&md,sizeof(unsigned int)*2);
			break;
		}
		case SWORDFISH_RANK: {
			md.param1=clientid;
			write(Clients[clientid].fd,&md,sizeof(unsigned int)*2);
			break;
		} 
		case SWORDFISH_CONSUMETIME: {
			write(Clients[clientid].fd,&md,sizeof(unsigned int));
			break;
		}
		case SWORDFISH_SEND: {
			mini_Packet m_packet=mini_Packet(md.param1, clientid, md.param3 ,md.param2);
			m_packet.data = (char *)malloc(md.param3);
			all_read(Clients[clientid].fd,m_packet.data,md.param3);
			Clients[md.param1].incomming_packets.push_back(m_packet);
			md.param1=0; //FIXME
			write(Clients[clientid].fd,&md,sizeof(unsigned int)*2);
			HandleClient(md.param1);
			break;
		}
		case SWORDFISH_RECV: {
			//printf("SWORDFISH_RECV[%u]: recv_buffer=%u\n",clientid,md.param1);
			Clients[clientid].state=RECV_WAIT;
			HandleClient(clientid);
			break;
		}
		case SWORDFISH_PROBE: {
			if(Clients[clientid].incomming_packets.size() > 0) {
				md.param1=Clients[clientid].incomming_packets.front().from;
				md.param2=Clients[clientid].incomming_packets.front().tag;
				md.param3=Clients[clientid].incomming_packets.front().size;
				md.param4=1; //FIXME
			} else {
				md.param4=0;
			}
			write(Clients[clientid].fd,&md,sizeof(unsigned int)*5);
			break;
		}
		default: {
			std::cout << "can't fullfill request=" << md.request << std::endl;
			exit(1);
		}
	}
}
/* }}} */

/* {{{ NetworkThread */
void NetworkThread()
{
	struct epoll_event events[100];
	int epfd = epoll_create(10);

	for(int i = 0; i < CLIENTS_NUMB; i++) {
		events[i].events = EPOLLIN | EPOLLPRI | EPOLLERR | EPOLLHUP;
		events[i].data.u32 = i;
		int res = epoll_ctl(epfd, EPOLL_CTL_ADD, Clients[i].fd, &events[i]);
		if(res!=0) {
			std::cout << "ERROR: epoll_ctl wasn't successfull" << std::endl;
			exit(1);
		}
	}

	std::cout << "Hello from NetworkThread\n";
	while(1) {
		struct epoll_event evs[100];
		Worker();
		int nfds = epoll_wait(epfd, evs, 5, 100);
		if (nfds < 0) {
			std::cout << "ERROR in epoll_wait!" << std::endl;
			exit(1);
		}
		for(int i = 0; i < nfds; i++) {
			HandleMsg(evs[i].data.u32);
		}
		Worker();
	}
}
/* }}} */

/* {{{ main */
int main(int argc, char *argv[])
{
	struct sockaddr_in servaddr, cliaddr;
	int netfd,on;
	
	if(argc!=2) {
		std::cout << "Usage: simpel <number of nodes>" << std::endl;
		exit(1);
	}
	CLIENTS_NUMB = atoi(argv[1]);
	std::cout << "Number of nodes: " << CLIENTS_NUMB << std::endl;

	Clients = new ClientControlBlock[CLIENTS_NUMB];

	for(int i=0;i<CLIENTS_NUMB;i++)
		Clients[i].state=NOTHING;

	if( (netfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		std::cout << "creating a socket failed!\n";
		exit(1);
	}
	bzero(&servaddr, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port = htons(17823);
	setsockopt(netfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	if( bind(netfd, (struct sockaddr *) &servaddr, sizeof(servaddr)) == -1) {
		printf("%s\n","bind failed!");
		puts(strerror(errno));
		exit(1);
	}
	listen(netfd, 10); 
	socklen_t clilen = sizeof(cliaddr);
	printf("Server socket created (port=%d)!\n",17823);
	int server_set_rank=0;

	for(int i = 0; i < CLIENTS_NUMB; i++) {
            int tmp_fd = accept(netfd, (struct sockaddr *) &cliaddr, &clilen);
            int thisRank;
            read(tmp_fd, (void*)&thisRank,4);
            if(thisRank==99999) {
				if(server_set_rank!=1 && i!=0) {
					std::cout << "can't mix server and client set rank!" << std::endl;
					exit(1);
				}
				server_set_rank=1;
				Clients[i].fd = tmp_fd;
				thisRank=i;
            } else {
				if(server_set_rank==1) {
					std::cout << "can't mix server and client set rank!" << std::endl;
					exit(1);
				}
				if(thisRank>=CLIENTS_NUMB) {
					std::cout << "rank to high!\n" << std::endl;
					exit(1);
				}
				if(Clients[thisRank].fd!=0) {
					std::cout << "a clients was allready connected with this rank (" << thisRank << ")" << std::endl;
					exit(1);
				}
            	Clients[thisRank].fd = tmp_fd;
			}
            std::cout << "Accepted a connection for a client with rank: " << thisRank << "\n";
        }

		NetworkThread();
	return 0;
}
/* }}} */
