/** Christian Leber <christian@leber.de>
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <string.h>
#include <sys/uio.h>

#include <time.h>
#include <pthread.h>

#define UINT_SIZE 4

#include "swf_client.h"
time_t stime_;

ssize_t swfcom_read_(swf_conn *con, void *buf, size_t count)
{
	ssize_t res;
	res = read(con->fd,buf,count);
	if(res<0) {
		printf("SWFCOM_ERROR: read error\n");
		exit(1);
	}
	return res;
}

ssize_t swfcom_write_(swf_conn *con, const void *buf, size_t count)
{
	ssize_t res;
	res = write(con->fd,buf,count);
	if(res<count) {
		printf("SWFCOM_ERROR: write did not succeed\n");
		exit(1);
	}
	return res;
}

ssize_t swfcom_read(swf_conn *con, void *buf, size_t count)
{
	ssize_t bytes_done=0, res, to_read;
	while(bytes_done < count) {
		if( (count-bytes_done) >16000) {
			to_read=16000;
		} else {
			to_read=count-bytes_done;
		}
		res = read(con->fd, buf+bytes_done,count);// to_read);
	//FIXME	
		if(res<=0) {
			printf("SWFCOM_ERROR: swfcom_read: count=%i res=%i\n",
				count,res);
			exit(1);
		}
		bytes_done+=res;
	}
	return bytes_done;
}
ssize_t swfcom_write(swf_conn *con, const void *buf, size_t count)
{
	ssize_t bytes_done=0, res, to_write;
	while(bytes_done < count) {
		if( (count-bytes_done) >16000) {
			to_write=16000;
		} else {
			to_write=count-bytes_done;
		}
		res = write(con->fd, buf+bytes_done, count);//to_write);
		if(res<=0) {
			printf("SWFCOM_ERROR: write did not succeed\n");
			exit(1);
		}
		bytes_done+=res;
	}
	return bytes_done;
}

int swfcom_readv(swf_conn *con, unsigned int src, const struct iovec *vector,
               size_t count)
{
	int i;
	unsigned int from,tag,size;
	printf("would read ");
	for(i=0; i<count; i++) {
		
		printf("%i ",vector[i].iov_len);
	}	
	printf("bytes from swf\n");
		swfcom_send_probe_request(con,&from,&tag,&size);
		printf("swfcom_readv: size=%i\n",size);
	return 0;
}

int swfcom_writev(swf_conn *con, unsigned int dest, const struct iovec *vector,
		size_t count)
{
	int i,size=0;
	unsigned int res;
	char *buffer;
	printf("would write ");
	for(i=0; i<count; i++) {
		printf("%i ",vector[i].iov_len);
		size += vector[i].iov_len;		
	}	
	printf(" = %i bytes to swf\n",size);
	buffer = (char *)malloc(size);
	res=swfcom_send(con,dest,42,size,buffer);
	free(buffer);
	printf("wrote %i\n",res);
	swfcom_consumetime(con,10000);
	printf("returned from consumetime\n");
	return 0;
}

int swfcom_buffer_add(swf_conn *con,char *addthis, unsigned int addsize)
{
	return 0;
}

/* throws amount bytes away from the buffer */
int swfcom_buffer_throw_away_bytes(swf_conn *con,char amount)
{
	return 0;
}

int Tcpstuff_Get_Socket(char *host, char *port)
{
	int output_socket; /* we'll return this */
	int valuu;
	struct addrinfo *hints, *res, *ressave;

	hints = malloc(sizeof(struct addrinfo));
	bzero(hints, sizeof(struct addrinfo));
	
	hints->ai_family = AF_UNSPEC;
	hints->ai_socktype = SOCK_STREAM;
	
	valuu = getaddrinfo(host, port, hints, &res);
	if( valuu != 0) {
		printf("Tcpstuff_Get_Socket: %s\n", gai_strerror(valuu));
		exit(1);
	}
	ressave=res;
	do {
		output_socket = socket(res->ai_family, res->ai_socktype,
				res->ai_protocol);
		if(output_socket < 0) continue;
		if(connect(output_socket, res->ai_addr, res->ai_addrlen) == 0)
			break;
		close(output_socket);
	} while( ( res=res->ai_next) != NULL);
	
	if(res==NULL) {
		printf("fatal error: Can't connect to %s:%s!\n", host, port);
		exit(1);
	}
	freeaddrinfo(ressave);
	printf("Tcpstuff_Get_Socket: fd=%i\n",output_socket);
	return (output_socket);	
}

void swfcom_init_struct(swf_conn *con)
{
	con->is_connected=0;	
}

void swfcom_connect_with_rank(swf_conn *con,unsigned int rank)
{
	if(con->is_connected==0) {
		printf("swfcom_connect_with_rank: %i\n",rank);
		con->rank=rank;
		swfcom_write(con,&con->rank,4);
		con->is_connected=1;
		swfcom_consumetime(con,200);
		swfcom_get_rank(con);
		con->buffer=NULL;
		con->buffer_size=0;
		con->buffer_length=0;
	}	
}

/* SWORDFISH_FINISH */
void swfcom_finish(swf_conn *con)
{
	struct DataMessage m={1,0,0,0,0,0};
	swfcom_write(con,&m,UINT_SIZE*6);
	close(con->fd);
}

/* SWORDFISH_SIZE */
void swfcom_get_size(swf_conn *con)
{
	struct DataMessage m={2,0,0,0,0,0};
	swfcom_write(con,&m,UINT_SIZE*6);
	swfcom_read(con,&m,UINT_SIZE*2);
	printf("swfcom_get_size: %u\n",m.param1);
	con->network_size=m.param1;
}

/* SWORDFISH_RANK */
void swfcom_get_rank(swf_conn *con)
{
	struct DataMessage m={3,0,0,0,0,0};
	swfcom_write(con,&m,UINT_SIZE*6);
	swfcom_read(con,&m,UINT_SIZE*2);
	printf("swfcom_get_rank: %u fd: %i\n",m.param1,con->fd);
	con->rank = m.param1;
	return;
}

/* SWORDFISH_TIME */
unsigned long swfcom_get_time(swf_conn *con)
{
	struct DataMessage m={4,0,0,0,0,0};
	swfcom_write(con,&m,UINT_SIZE*6);
	swfcom_read(con,&m,sizeof(unsigned int) *2);
	return m.param1;
}

/* SWORDFISH_CONSUMETIME */
//FIXME data type has to be fixed, unsigned long is on AMD64 NOT 32 bit long */
//just on windows...
unsigned int swfcom_consumetime(swf_conn *con, unsigned int time)
{
	struct DataMessage m={5,time,0,0,0,0};
	swfcom_write(con,&m,UINT_SIZE*6);
	swfcom_read(con,&m,UINT_SIZE*1);
	return m.param1;
}

/* SWORDFISH_SEND */
unsigned swfcom_send(swf_conn *con, unsigned dest,unsigned tag,
		unsigned size, char* buffer)
{
	struct DataMessage m={6,dest,tag,size,0,0};

//	if(dest == con.rank) {

//	} else {
		swfcom_write(con,&m,UINT_SIZE*6);
		swfcom_write(con,buffer,size);
	
		swfcom_read(con,&m,UINT_SIZE*2);
		return m.param1;
//	}
}

/* SWORDFISH_RECV */
void swfcom_recv(swf_conn *con,unsigned *from, unsigned* tag, unsigned* size,
		char* buffer, unsigned int length)
{
	struct DataMessage m={7,length,0,0,0,0};
	swfcom_write(con,&m,UINT_SIZE*6);
///	printf("[swf%u] swfcom_recv 1\n",con->rank);
	
	swfcom_read(con,&m,UINT_SIZE*4);
///        printf("[swf%u] swfcom_recv 2\n",con->rank);
	
	*from=m.param1;
	*tag=m.param2;
	*size=m.param3;
	swfcom_read(con,buffer,*size);
///        printf("[swf%u] swfcom_recv 3\n",con->rank);
}

unsigned swfcom_send_probe_request(swf_conn *con,unsigned* from,
	       	unsigned* tag, unsigned*size)
{
	struct DataMessage m={10,0,0,0,0,100}; //FIXME 100!
	swfcom_write(con,&m,UINT_SIZE*6);

	swfcom_read(con,&m,sizeof(unsigned int)*5);
	*from=m.param1;
	*tag=m.param2;
	*size=m.param3;
	return m.param4;
}


